﻿using System;
using System.Collections.Generic;
using System.Web;
using System.Xml;
using System.Web.Caching;
using System.Configuration;
using System.Data;
using System.IO;

namespace CacheMgr
{
    /// <summary>
    /// Summary description for CacheManager
    /// </summary>
    // TODO: Make this class as singleton
    [Serializable]
    public class CacheManager
    {

        LogFile.Log logobj = new LogFile.Log();
        Custom customobj = Custom.Instance;
        string errorLog = string.Empty;
        private string structHTML = string.Empty;
        private string structScript = string.Empty;
        private string structBtns = string.Empty;
        private string taskBtns = string.Empty;
        private string structCaption = string.Empty;
        private string structName = string.Empty;
        private string strSubmitCancel = string.Empty;
        private string attachHTML = string.Empty;
        private string tstHeaderHTML = string.Empty;
        private static string toolbarBtnIcons = string.Empty;
        private Boolean IsTstObjNew = true;    // the value needs to be false in production. 
        private int NoOfDaysForExpiry = 1;
        Util.Util util = new Util.Util();
        Custom cusObj = Custom.Instance;

        //static RedisServer fdwObj = RedisServer.Instance;
        FDW fdwObj = FDW.Instance;


        public string StructureHtml
        {
            get { return structHTML; }
            set { structHTML = value; }
        }

        public string TstHeaderHtml
        {
            get { return tstHeaderHTML; }
            set { tstHeaderHTML = value; }
        }

        public string ToolbarBtnIcons
        {
            get { return toolbarBtnIcons; }
            set { toolbarBtnIcons = value; }
        }

        public string AttachmentHtml
        {
            get { return attachHTML; }
            set { attachHTML = value; }
        }

        public string StructureScript
        {
            get { return structScript; }
            set { structScript = value; }
        }

        public string StructureBtns
        {
            get { return structBtns; }
            set { structBtns = value; }
        }

        public string TaskButtons
        {
            get { return taskBtns; }
            set { taskBtns = value; }
        }

        public string StructureCaption
        {
            get { return structCaption; }
            set { structCaption = value; }
        }

        public string StructureName
        {
            get { return structName; }
            set { structName = value; }
        }

        public string StructureSubmitCancel
        {
            get { return strSubmitCancel; }
            set { strSubmitCancel = value; }
        }

        public Boolean IsTstructNew
        {
            get { return IsTstObjNew; }
            set { IsTstObjNew = value; }
        }

        public CacheManager(string errorLog)
        {
            this.errorLog = errorLog;
        }

        /// <summary>
        /// Function to get the TStruct definition object from Cache.
        /// If it exist in cache then check if the structure was changed, if it has not changed then get it from cache.
        /// else get the structure from the db.
        /// </summary>
        /// <param name="Proj">Project name</param>
        /// <param name="Sid">Session Id</param>
        /// <param name="User">User name</param>
        /// <param name="TransId">Transaction Id</param>
        /// <param name="AxRole">Role</param>
        /// <returns>struct def object.</returns>
        public TStructDef GetStructDef(string projectName, string sessionID, string userName, string transId, string AxRole)
        {

            TStructDef strObj;
            string fdKey = Constants.REDISTSTRUCT;
            string newTid = transId;
            if (transId.EndsWith("-design"))
            {
                transId = transId.Substring(0, transId.IndexOf("-"));
                fdKey = Constants.REDISTSTRUCTDESIGN;
                newTid = transId + "-design";
            }
            string fileName = "opentstruct-" + transId;
            string language = HttpContext.Current.Session["language"].ToString();
            FDR fObj = (FDR)HttpContext.Current.Session["FDR"];
            logobj.CreateLog("Get Struct from Cache Manager. User: " + userName + " Role: " + AxRole, sessionID, fileName, "");
            strObj = (TStructDef)fObj.TstructDefFromRedis(util.GetRedisServerkey(fdKey));
            if (strObj != null && !IsStructureUpdated(strObj, transId, sessionID, projectName, userName))
            {
                //TODO- Fastdata               
                strObj.IsObjFromCache = true;
                logobj.CreateLog("Getting Structure from Redis " + userName + " Role: " + AxRole, sessionID, fileName, "");
            }
            else
            {
                logobj.CreateLog("Getting Structure from DB " + userName + " Role: " + AxRole, sessionID, fileName, "");
                strObj = GetStructDefFromDb(newTid, AxRole, projectName, userName, sessionID);
            }

            var memFastDataDef = fdwObj.DSasJsonFromRedis(Constants.RS_FD_DEFINITION);
            if (memFastDataDef != null)
                strObj.dsFastDataDef = memFastDataDef;
            return strObj;

        }

        /// <summary>
        /// Function to get the struct def object from the data base.
        /// </summary>
        /// <param name="TransId">Transaction Id</param>
        /// <param name="AxRole">Role</param>
        /// <param name="strObj">Struct def object</param>
        /// <returns>struct def object.</returns>
        public TStructDef GetStructDefFromDb(string transId, string AxRole, string proj, string axpUser, string sessionID)
        {

            string fdKey = Constants.REDISTSTRUCT;
            if (transId.EndsWith("-design"))
            {
                transId = transId.Substring(0, transId.IndexOf("-"));
                fdKey = Constants.REDISTSTRUCTDESIGN;
            }
            string res = string.Empty, fileName = "opentstruct-" + transId;
            string iXml = string.Empty;
            string language = HttpContext.Current.Session["language"].ToString();
            logobj.CreateLog("    Calling webservice to get Structure from DB", sessionID, fileName, "");
            ASBExt.WebServiceExt objWebServiceExt = new ASBExt.WebServiceExt();
            iXml = iXml + "<root axpapp='" + proj + "' sessionid='" + sessionID + "' axdesign='true' trace='" + errorLog + "' ><sname>tstructs</sname><name>" + transId + "</name><caption></caption>";
            iXml = iXml + HttpContext.Current.Session["axApps"].ToString() + HttpContext.Current.Application["axProps"].ToString() + "</root>";

            //Call service
            res = objWebServiceExt.CallGetStructureWS(transId, iXml);

            if (res.Contains(Constants.ERROR) == true)
            {
                res = res.Replace(Constants.ERROR, "");
                res = res.Replace("</error>", "");
                res = res.Replace("\n", "");
                throw (new Exception(res));
            }
            else
            {
                //GetTstruct Custom details
                DataTable srchCols = new DataTable();
                try
                {
                    // srchCols = GetTstructCustoms(transId); // Avoiding getchoice call and same result merged in getstruct XML as node

                    XmlDocument xmlDoc = new XmlDocument();
                    xmlDoc.LoadXml(res);
                    string customDetails = string.Empty;
                    XmlNodeList customDetailsXML = xmlDoc.GetElementsByTagName("axpconfigs_" + transId);
                    if (customDetailsXML[0] != null)
                        customDetails = customDetailsXML[0].OuterXml;
                    srchCols = GetTstructCustoms(transId, customDetails);
                }
                catch (Exception ex)
                {
                    throw ex;
                }

                DataSet dsRapidDef = new DataSet();

                if (ConfigurationManager.AppSettings["IsRapidLoad"] != null && Convert.ToBoolean(ConfigurationManager.AppSettings["IsRapidLoad"].ToString()) == true && (HttpContext.Current.Session["AxIsPerfCode"] != null && HttpContext.Current.Session["AxIsPerfCode"].ToString().ToLower() == "true"))
                {
                    dsRapidDef = GetRapidDefinition(transId);
                }

                TStructDef strObj = new TStructDef(res, srchCols, dsRapidDef);

                //this call removed now if its required need to analyse 
                //Getting Hint text for particular Tstruct
                //ASB.WebService objWeb = new ASB.WebService();
                //try
                //{
                //    strObj.HlpText = objWeb.GetHelpText(transId);
                //}
                //catch (Exception ex)
                //{
                //    throw ex;
                //}

                //DMS Integration
                try
                {
                    strObj = customobj.AxAfterGetStruct(res, transId, strObj);
                }
                catch (Exception ex)
                {
                    throw ex;
                }

                strObj.IsObjFromCache = false;
                strObj.structRes = res;
                IsTstObjNew = true;

                //TODO: REDIS
                fdwObj.SaveInRedisServer(util.GetRedisServerkey(fdKey), strObj, Constants.REDISTSTRUCT);

                if (strObj.axdesignJson != "")
                {
                    string fdKeyaxDesign = Constants.REDISTSTRUCTAXDESIGN;
                    fdwObj.SaveInRedisServer(util.GetRedisServerkey(fdKeyaxDesign), strObj.axdesignJson, Constants.REDISTSTRUCTAXDESIGN);
                }

                if (strObj.dcTable.Count > 0 && strObj.dcTable[0].ToString() != "")
                {
                    string fdKeydcTable = Constants.REDISTSTRUCTTABLE;
                    fdwObj.SaveInRedisServer(util.GetRedisServerkey(fdKeydcTable), strObj.dcTable[0], Constants.REDISTSTRUCTTABLE);
                }
                logobj.CreateLog("    Adding the structure to cache. Updated time-" + strObj.tstUpdateOn, sessionID, "GetCacheDetails", "");
                logobj.CreateLog("    Returning the structure.", sessionID, fileName, "");
                return strObj;
            }
        }

        private DataSet GetRapidDefinition(string transId)
        {
            DataSet dsDef = new DataSet();

            try
            {
                DBContext objDB = new DBContext();
                dsDef = objDB.GetRapidDefData(Constants.GET_RAPID_DEF_SP, transId);
                logobj.CreateLog("    Data from Rapid definition.-" + dsDef, HttpContext.Current.Session.SessionID, "RapidDefinition-" + transId, "");
            }
            catch (Exception ex)
            {
                logobj.CreateLog("Exception in GetRapidDefinition-" + ex.Message, HttpContext.Current.Session.SessionID, "ExcRapidDefinition-" + transId, "new");
                //throw ex;
            }

            return dsDef;
        }

        private DataTable GetTstructCustoms(string transId, string cutResult)
        {
            string cutTable = "axpconfigs_" + transId;
            string srchCols = string.Empty;
            DataSet ds = new DataSet();
            if (cutResult != null || cutResult != "")
            {
                if (cutResult.Contains(Constants.ERROR) == true)
                {
                    return new DataTable();
                }
                if (cutResult != string.Empty)
                {
                    try
                    {
                        StringReader sr = new StringReader(cutResult);
                        XmlTextReader reader = new XmlTextReader(sr);
                        ds.ReadXml(reader);
                    }
                    catch (Exception ex)
                    {
                        throw ex;
                    }
                }
                return ds.Tables[cutTable];
            }
            else
            {
                return new DataTable();
            }

            #region Before concurrent user 
            //string result = string.Empty;
            //string srchCols = string.Empty;
            //DataSet ds = new DataSet();
            //ASBCustom.CustomWebservice objWeb = new ASBCustom.CustomWebservice();
            //try
            //{
            //    string sqlStr = Constants.GET_TST_CUSTOM_ACTIONS.Replace(Constants.VAR_TRANSID, transId);
            //    result = objWeb.GetChoices(transId, sqlStr);
            //}
            //catch (Exception ex)
            //{
            //    logobj.CreateLog("Exception in GetTstructCustoms-" + ex.Message, HttpContext.Current.Session.SessionID, "GetTstructCustoms", "");
            //    throw ex;
            //}

            //if (result.Contains(Constants.ERROR) == true)
            //{
            //    return new DataTable();
            //}
            //if (result != string.Empty)
            //{
            //    try
            //    {
            //        StringReader sr = new StringReader(result);
            //        XmlTextReader reader = new XmlTextReader(sr);
            //        ds.ReadXml(reader);
            //    }
            //    catch (Exception ex)
            //    {
            //        throw ex;
            //    }
            //}
            //return ds.Tables["row"];
            #endregion
        }

        /// <summary>
        /// Function to check if the stucture definition was changed in the database by comparing the updatedOn and the updated date stored in the struct def object.
        /// </summary>
        /// <param name="strObj">Struct def object</param>
        /// <param name="TransId">Transaction Id</param>
        /// <returns>True if the structure was updated else false.</returns>
        public Boolean IsStructureUpdated(TStructDef strObj, string transId, string sessionID, string proj, string axpUser)
        {
            string isDevInstance = string.Empty;
            if (HttpContext.Current.Session["AxDevInstance"] != null)
                isDevInstance = HttpContext.Current.Session["AxDevInstance"].ToString();
            if (isDevInstance == "true")
            {
                System.Globalization.CultureInfo culture = default(System.Globalization.CultureInfo);
                culture = new System.Globalization.CultureInfo("en-GB");

                string dbTimeSql = string.Empty, dbTimeRes = string.Empty, fileName = string.Empty;
                ASBExt.WebServiceExt objWebServiceExt = new ASBExt.WebServiceExt();
                if (transId.ToLower().Equals("axglo"))
                    dbTimeSql = "<sqlresultset axpapp='" + proj + "' sessionid='" + sessionID + "' trace='" + errorLog + "'><sql>select updatedon upd from tstructs where lower(name) ='" + transId.ToLower() + "' and blobno=1</sql>";
                else
                    dbTimeSql = "<sqlresultset axpapp='" + proj + "' sessionid='" + sessionID + "' trace='" + errorLog + "'><sql>select updatedon upd from tstructs where name ='" + transId + "' and blobno=1</sql>";

                dbTimeSql += HttpContext.Current.Session["axApps"].ToString() + HttpContext.Current.Application["axProps"].ToString() + "</sqlresultset>";
                fileName = "opentstruct-" + transId;
                logobj.CreateLog("Calling DB to check structure is updated.", sessionID, fileName, "");
                //Call service
                dbTimeRes = objWebServiceExt.CallGetChoiceWS(transId, dbTimeSql);

                if (dbTimeRes.Contains(Constants.ERROR) == true)
                {
                    dbTimeRes = dbTimeRes.Replace(Constants.ERROR, "");
                    dbTimeRes = dbTimeRes.Replace("</error>", "");
                    dbTimeRes = dbTimeRes.Replace("\n", "");
                    throw (new Exception(dbTimeRes));
                }
                else
                {
                    string dbResult = dbTimeRes.ToUpper();
                    XmlDocument xmlDoc = new XmlDocument();
                    XmlNode dbUpdatedtime = null;
                    xmlDoc.LoadXml(dbResult);
                    dbUpdatedtime = xmlDoc.SelectSingleNode("//RESPONSE/ROW/UPD");

                    string dbUpdDate = dbUpdatedtime.InnerText;
                    DateTime dbDate;
                    if (HttpContext.Current.Session["axdb"] != null)
                    {
                        if (HttpContext.Current.Session["axdb"].ToString().ToLower() == "ms sql")
                            culture = new System.Globalization.CultureInfo("en-US");
                    }


                    try
                    {
                        dbDate = DateTime.Parse(dbUpdDate, culture);
                    }
                    catch (System.FormatException ex)
                    {

                        dbDate = DateTime.Parse(dbUpdDate);
                    }

                    DateTime cacheTime;
                    string objTime = strObj.tstUpdateOn;
                    if (object.ReferenceEquals(objTime, ""))
                        objTime = "01/01/2000 00:00:01 AM";

                    try
                    {
                        cacheTime = DateTime.Parse(objTime, culture);
                    }
                    catch (System.FormatException ex)
                    {
                        cacheTime = DateTime.Parse(objTime);
                    }

                    logobj.CreateLog("    Cache time is " + cacheTime + ", DB time is " + dbDate + ", transid-" + transId, sessionID, "GetCacheDetails", "");
                    return dbDate > cacheTime;
                }
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Function to get the Structure HTML from cache.
        /// </summary>
        /// <param name="transId">Transaction Id</param>
        /// <param name="AxRole">Role</param>
        public void GetStructureHTML(string transId, string AxRole, string sessionID, string language)
        {
            string cachName = string.Empty, fileName = string.Empty;
            string fdKey = Constants.REDISTSTRUCT;
            if (transId.EndsWith("-design"))
            {
                transId = transId.Substring(0, transId.IndexOf("-"));
                fdKey = Constants.REDISTSTRUCTDESIGN;
            }
            fileName = "opentstruct-" + transId;
            logobj.CreateLog("    Getting structure from Cache.", sessionID, fileName, "");
            string proj = HttpContext.Current.Session["project"].ToString();
            string cacheKey = (util.GetRedisServerkey(fdKey));
            FDR fObj = (FDR)HttpContext.Current.Session["FDR"];
            DataTable dt = fObj.DTasJsonFromRedis("TSTDTLS-" + cacheKey);

            if (dt != null)
            {
                if (dt.Rows.Count > 0)
                {
                    structHTML = dt.Rows[0][Constants.structHTML].ToString();
                    tstHeaderHTML = dt.Rows[0][Constants.tstHeaderHTML].ToString();
                    toolbarBtnIcons = dt.Rows[0][Constants.toolbarBtnIcons].ToString();
                    attachHTML = dt.Rows[0][Constants.attachHTML].ToString();
                    taskBtns = dt.Rows[0][Constants.taskBtns].ToString();
                    structScript = dt.Rows[0][Constants.structScript].ToString();
                    structBtns = dt.Rows[0][Constants.structBtns].ToString();
                    structCaption = dt.Rows[0][Constants.structCaption].ToString();
                    strSubmitCancel = dt.Rows[0][Constants.strSubmitCancel].ToString();
                }
            }
        }

        /// <summary>
        /// Function to store the Structure HTML into cache.
        /// </summary>
        /// <param name="transId">Transaction Id</param>
        /// <param name="AxRole">Role</param>
        public void SetStructureHTML(string transId, string AxRole, string language)
        {

            string fdKey = Constants.REDISTSTRUCT;
            if (transId.EndsWith("-design"))
            {
                transId = transId.Substring(0, transId.IndexOf("-"));
                fdKey = Constants.REDISTSTRUCTDESIGN;
            }
            string proj = HttpContext.Current.Session["project"].ToString();
            string cacheKey = (util.GetRedisServerkey(fdKey));
            string cachName = string.Empty;

            DataTable dt = new DataTable();
            dt.Columns.Add(Constants.structHTML);
            dt.Columns.Add(Constants.tstHeaderHTML);
            dt.Columns.Add(Constants.toolbarBtnIcons);
            dt.Columns.Add(Constants.attachHTML);
            dt.Columns.Add(Constants.taskBtns);
            dt.Columns.Add(Constants.structScript);
            dt.Columns.Add(Constants.structBtns);
            dt.Columns.Add(Constants.structCaption);
            dt.Columns.Add(Constants.strSubmitCancel);

            DataRow dr = dt.NewRow();
            dr[Constants.structHTML] = structHTML;
            dr[Constants.tstHeaderHTML] = tstHeaderHTML;
            dr[Constants.toolbarBtnIcons] = toolbarBtnIcons;
            dr[Constants.attachHTML] = attachHTML;
            dr[Constants.taskBtns] = taskBtns;
            dr[Constants.structScript] = structScript;
            dr[Constants.structBtns] = structBtns;
            dr[Constants.structCaption] = structCaption;
            dr[Constants.strSubmitCancel] = strSubmitCancel;
            dt.Rows.Add(dr);

            fdwObj.SaveInRedisServerDT("TSTDTLS-" + cacheKey, dt, Constants.REDISTSTRUCT);
        }
    }
}
